# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-15"
# Created: 2017-04-15 16:09

from .basic_msg import BasicMessage


class InputMessage(BasicMessage):
    """
    Input message sent by communicator
    """

    def __init__(self,
            uuid=None, timestamp=None,
            source=None, data=None, metadata=None
    ):
        super(InputMessage, self).__init__(uuid, timestamp)
        self.source = source
        """ Communicator sending this message
            :type : unicode """
        self.data = data
        """ Data received by communicator (Depends on communicator)
            :type : object """
        self.metadata = metadata
        """ Meta information added by communicator (Depends on communicator)
            :type : object """



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.4"
__date__ = "2017-04-23"
# Created: 2017-04-17 20:26

from . import dto
from . import reactor
from . import manager
from .reactor import ReactorModule
from .reactor.events import setup_kombu
from .manager import IntentManager, ActorManager

__all_ = ["dto", "reactor", "IntentManager", "ActorManager", "setup_kombu"]



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.1"
__date__ = "2017-04-20"
# Created: 2017-04-15 14:06

from abc import ABCMeta, abstractmethod
import datetime
import uuid

from flotils import Loadable, StartStopable
from kombu import producers, Connection, Exchange

from .events import EventListener
from ..dto import InputMessage, IntentMessage, ActorMessage


class ReactorModule(Loadable, StartStopable):
    __metaclass__ = ABCMeta

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(ReactorModule, self).__init__(settings)
        nameko_sett = settings['nameko']
        broker = settings.get('broker', nameko_sett.get('AMQP_URI'))
        serializer = settings.get('serializer', nameko_sett.get('serializer'))
        nameko_sett.setdefault('AMQP_URI', broker)
        nameko_sett.setdefault('serializer', serializer)
        self._connection_timeout = settings.get('connection_timeout', 2.0)

        evt_sett = settings.get("event_listener", {})
        evt_sett.setdefault('AMQP_URI', broker)
        evt_sett.setdefault('accept_formats', [serializer])

        evt_sett_nameko = dict(evt_sett)
        evt_sett_nameko['AMQP_URI'] = nameko_sett['AMQP_URI']
        evt_sett_nameko.update(settings.get("event_listener_nameko", {}))

        self._event_listener = EventListener(evt_sett)
        self._event_listener.process_message = self.react
        self._event_listener_nameko = EventListener(evt_sett_nameko)
        self._event_listener_nameko.process_message = self.react_nameko
        self._connection = Connection(broker)
        self._com_exchanges = {}

    def class_name(self):
        return self.__class__.__name__

    def _register(self, exchange, routing_key, queue=None):
        if queue is None:
            queue = self.class_name()
        self._event_listener.register_event(
            exchange, "topic", queue, routing_key
        )

    def _register_nameko(self, pub_service_name, pub_event_type):
        self._event_listener_nameko.register_nameko_event(
            pub_service_name, pub_event_type, "topic",
            sub_service_name=self.class_name()
        )

    def get_exchange(self, key):
        if key not in self._com_exchanges:
            self._com_exchanges[key] = Exchange(
                key, type="topic", durable=True, auto_delete=False
            )
        return self._com_exchanges[key]

    def emit(self, message, exchange):
        """
        Emit data

        :param message: Message data to emit
        :type message: dto.InputMessage | dto.IntentMessage
        :param exchange: Exchange to publish message on
        :type exchange: kombu.Exchange
        :rtype: None
        """
        if isinstance(message, ActorMessage):
            routing_key = "say"
        elif isinstance(message, IntentMessage):
            routing_key = "intent_new"
        elif isinstance(message, InputMessage):
            routing_key = "input_new"
        else:
            self.error("No routing key for\n{}".format(message))
            raise ValueError("No routing key")
        if message.timestamp is None:
            message.timestamp = datetime.datetime.utcnow()
        if message.uuid is None:
            message.uuid = "{}".format(uuid.uuid4())
        # self.debug("{}:\n{}".format(routing_key, message))
        self._connection.ensure_connection()
        with producers[self._connection].acquire(
                block=True, timeout=self._connection_timeout
        ) as producer:
            producer.publish(
                message,
                compression="bzip2",
                exchange=exchange,
                declare=[exchange],
                retry=True,
                routing_key=routing_key,
                serializer="datetimejson"
            )

    @abstractmethod
    def react(self, exchange, routing_key, msg):
        """
        React to input

        :param exchange: What exchange the message came from
        :type exchange: unicode
        :param routing_key: What routing key was used
        :type routing_key: unicode
        :param msg: The received message
        :type msg: dict
        :return: Accept the input or refuse it
        :rtype: bool
        """
        raise NotImplementedError

    @abstractmethod
    def react_nameko(self, exchange, routing_key, msg):
        """
        React to nameko input

        :param exchange: What exchange the message came from
        :type exchange: unicode
        :param routing_key: What routing key was used
        :type routing_key: unicode
        :param msg: The received message
        :type msg: dict
        :return: Accept the input or refuse it
        :rtype: bool
        """
        raise NotImplementedError

    def start(self, blocking=False):
        self._event_listener.start(False)
        self._event_listener_nameko.start(False)
        super(ReactorModule, self).start(blocking)

    def stop(self):
        if not self._is_running:
            # Already stopped
            return
        try:
            self._event_listener.stop()
        except:
            self.exception("Failed to stop event listener")
        try:
            self._event_listener_nameko.stop()
        except:
            self.exception("Failed to stop event listener nameko")
        super(ReactorModule, self).stop()



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-02-19"
# Created: 2017-02-19 22:51


class ReactorException(Exception):
    pass


class NoProxyException(ReactorException):
    pass



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2016, Florian JUNG"
__license__ = "All rights reserved"
__version__ = "0.1.2"
__date__ = "2017-04-20"
# Created: 2016-08-19 19:41

import threading

from kombu import Exchange, Connection, Queue
from kombu.mixins import ConsumerMixin
from kombu.serialization import register
from flotils.runable import StartStopable
from flotils import Loadable


def setup_kombu():
    from ..dto.utils import encode, decode
    register(
        "datetimejson", encode, decode, "application/datetime-json", "utf-8"
    )


class EventListener(ConsumerMixin, Loadable, StartStopable):

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        Loadable.__init__(self, settings)
        StartStopable.__init__(self, settings)
        self._broker_url = settings['AMQP_URI']
        self._heartbeat = settings.get('heartbeat', None)
        events = settings.get('events', [])
        """ Which events to monitor
            :type : list[dict[unicode, unicode | callable]] """
        self._exchanges = {}
        """ Created exchanges
            :type : dict[unicode, kombu.Exchange] """
        self.process_message = settings.get('process_message')
        """ Process function
            :type : callable """
        self.connection = None
        """ Connection to broker
            :type : kombu.Connection """
        self._queues = []
        """ Queues listening on
            :type : list[kombu.Queue] """
        self._accept_formats = settings.get('accept_formats', ["json"])
        """ Which formats for serialization to support
            :type : list[unicode] """
        self.info("Registering events..")
        for event in events:
            if "pub_service_name" in event:
                self.debug("Registering as nameko event:\n{}".format(event))
                self.register_nameko_event(**event)
            elif "exchange" in event:
                self.debug("Registering as normal event:\n{}".format(event))
                self.register_event(**event)
            else:
                self.error("Could not register event:\n{}".format(event))

    def _thread_wrapper(self, func):
        """
        Wrap function for exception handling with threaded calls

        :param func: Function to call
        :type func: callable
        :rtype: None
        """
        try:
            func()
        except:
            self.exception("Threaded execution failed")

    def register_event(
            self, exchange_name, exchange_type, queue_name, routing_key,
            queue_durable=True,
            queue_auto_delete=False, exchange_auto_delete=False
    ):
        self.debug(
            "Registering Exchange({}, {})".format(exchange_name, exchange_type)
        )
        self.debug(
            "Registering Queue({}, {})".format(queue_name, routing_key)
        )
        if exchange_name not in self._exchanges:
            self._exchanges[exchange_name] = Exchange(
                exchange_name, type=exchange_type,
                auto_delete=exchange_auto_delete
            )
        exchange = self._exchanges[exchange_name]
        self._queues.append(Queue(
            queue_name, exchange, routing_key=routing_key
        ))
        self._queues[-1].durable = queue_durable
        self._queues[-1].auto_delete = queue_auto_delete

    def register_nameko_event(
            self,
            pub_service_name, pub_event_type, exchange_type="topic",
            sub_service_name=None, sub_handle_name="handle_event"
    ):
        self.register_event(
            "{}.events".format(pub_service_name), exchange_type,
            "evt-{}.{}--{}.{}".format(
                pub_service_name, pub_event_type,
                sub_service_name, sub_handle_name
            ), pub_event_type, exchange_auto_delete=True
        )

    def get_consumers(self, Consumer, channel):
        self.debug("()")
        return [Consumer(
            queues=self._queues,
            accept=self._accept_formats,
            callbacks=[self._process_message]
        )]

    def _process_message(self, body, message):
        exchange = message.delivery_info['exchange']
        routing_key = message.delivery_info['routing_key']

        if self.process_message:
            try:
                self.process_message(exchange, routing_key, body)
            except:
                self.exception("Failed to call process function")
        else:
            self.warning("No process function found")
        # TODO: Send occurred errors via message
        message.ack()

    def _run_worker(self):
        with Connection(self._broker_url, heartbeat=self._heartbeat) as conn:
            self.connection = conn
            self.run()

    def start(self, blocking=False):
        super(EventListener, self).start(False)
        if not blocking:
            try:
                a_thread = threading.Thread(
                    target=self._thread_wrapper,
                    args=(self._run_worker,)
                )
                a_thread.daemon = True
                a_thread.start()
            except:
                self.exception("Failed to run receive loop")
                self.stop()
                return
        else:
            self._run_worker()

    def stop(self):
        self.should_stop = True
        super(EventListener, self).stop()
        self._exchanges = {}
        self._queues = []



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# from __future__ import unicode_literals

__author__ = "d01 <Florian Jung>"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.2"
__date__ = "2017-04-20"
# Created: ?

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import sys
import os
import re


if sys.argv[-1] == "build":
    os.system("python setup.py clean sdist bdist bdist_egg bdist_wheel")


def get_version():
    """
    Parse the version information from the init file
    """
    version_file = os.path.join("alexander_fw", "__init__.py")
    initfile_lines = open(version_file, "rt").readlines()
    version_reg = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(version_reg, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError(
        u"Unable to find version string in {}".format(version_file)
    )


def get_file(path):
    with open(path, "r") as f:
        return f.read()


version = get_version()
# readme = get_file("README.rst")
# history = get_file("HISTORY.rst")
requirements = get_file("requirements.txt").split("\n")

assert version is not None
# assert readme is not None
# assert history is not None
assert requirements is not None

setup(
    name="alexander_fw",
    version=version,
    description="",
    # long_description=readme + "\n\n" + history,
    author="the01",
    author_email="jungflor@gmail.com",
    url="https://github.com/the01/python-alexander",
    packages=[
        "alexander_fw",
        "alexander_fw.dto",
        "alexander_fw.manager",
        "alexander_fw.reactor"
    ],
    install_requires=requirements,
    license="MIT License",
    keywords="",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
    ]
)
